#!/usr/bin/env node

/*jslint node:true,indent:4*/

(function () {
    "use strict";

    var ArgParser = require('argparse').ArgumentParser,
        async = require('async'),
        pack = require('../pack'),
        minify = require('../minify-batch'),
        buildDoc = require('../builddoc'),
        path = require('path'),
        fs = require('fs'),
        commons = require('../commons'),
        perl = require('../perl'),
        os = require('os'),
        check = require('../check-testcases'),

        parser = new ArgParser({
            version: '0.4',
            addHelp: true,
            description: 'ePages 6 node.js helper'
        }),
        subparsers = parser.addSubparsers({
            title: 'subcommands',
            dest: 'subcommand'
        }),
        subparser = {
            minifier: subparsers.addParser(
                'minify',
                {
                    addHelp: true,
                    description: "Minify all JavaScript and CSS files within the lib folder.",
                    epilog: "Example: epnode minify -t Store -v 6.17.0 -f /**/*.ca.json"
                }
            ),
            pack: subparsers.addParser(
                'pack',
                {
                    addHelp: true,
                    description: "Create Packages for JavaScript and CSS files from the specified profile.",
                    epilog: "Example: epnode pack --js --css -p sf -p mobilesf -p editor -p testsuite -t Store -v 6.17.0"
                }
            ),
            builddoc: subparsers.addParser(
                'builddoc',
                {
                    addHelp: true,
                    description: "Create the jsduck documentation.",
                    epilog: "Example: epnode builddoc -i sources.json -o ./"
                }
            ),
            testcases: subparsers.addParser(
                'testcases',
                {
                    addHelp: true,
                    description: "Check for the presence of testcases",
                    epilog: "Example: epnode testcases -c jquery"
                }
            )
        },
        startTime = new Date().getTime(),
        buffer = '',
        args = {},
        parseArgs,
        subcommand,
        uid,
        gid,
        logPath;


    parser.addArgument(
        ['-l', '--extended-log'],
        {
            nargs: 0,
            required: false,
            help: 'Activate extended logging and stack tracing'
        }
    );

    subparser.minifier.addArgument(
        ['-v', '--store-version'],
        {
            nargs: 1,
            required: true,
            help: 'Store version e.g. 6.17.0'
        }
    );
    subparser.minifier.addArgument(
        ['-t', '--store-type'],
        {
            nargs: 1,
            required: true,
            help: 'Store type e.g. Store'
        }
    );
    subparser.minifier.addArgument(
        ['-w', '--webroot'],
        {
            nargs: 1,
            required: false,
            help: 'Epages webroot defaults to $EPAGES_WEBROOT'
        }
    );
    subparser.minifier.addArgument(
        ['-f', '--filter'],
        {
            nargs: 1,
            required: false,
            help: 'Filter files to be minified e.g. /**/*.js'
        }
    );

    subparser.pack.addArgument(
        ['-j', '--js'],
        {
            nargs: 0,
            required: false,
            help: "Create JavaScript package from profile ['bo', 'sf', 'mobilesf', 'editor', 'testsuite', 'sf_appointment']"
        }
    );
    subparser.pack.addArgument(
        ['-c', '--css'],
        {
            nargs: 0,
            required: false,
            help: "Create CSS package from profile ['sf', 'mobilesf', 'editor', 'testsuite']"
        }
    );
    subparser.pack.addArgument(
        ['-l', '--less'],
        {
            nargs: 0,
            required: false,
            help: "Create LESS package from profile ['bo', 'pageeditor']"
        }
    );
    subparser.pack.addArgument(
        ['-p', '--profile'],
        {
            nargs: 1,
            action: 'append',
            required: true,
            choices: ['bo', 'sf', 'mobilesf', 'editor', 'pageeditor', 'testsuite', 'sf_appointment'],
            help: 'Profile used for packaging'
        }
    );
    subparser.pack.addArgument(
        ['-t', '--store-type'],
        {
            nargs: 1,
            required: true,
            help: 'Store type e.g. Store'
        }
    );
    subparser.pack.addArgument(
        ['-v', '--store-version'],
        {
            nargs: 1,
            required: true,
            help: 'Store version e.g. 6.17.0'
        }
    );
    subparser.pack.addArgument(
        ['-w', '--webroot'],
        {
            nargs: 1,
            required: false,
            help: 'Epages webroot defaults to $EPAGES_WEBROOT'
        }
    );

    subparser.builddoc.addArgument(
        ['-i', '--include'],
        {
            nargs: 1,
            required: true,
            help: 'A JSON file which enumerates the files to include in the documentation.'
        }
    );
    subparser.builddoc.addArgument(
        ['-o', '--output'],
        {
            nargs: 1,
            required: false,
            help: 'The output folder for the documentation'
        }
    );
    subparser.builddoc.addArgument(
        ['-c', '--cartridge-folder'],
        {
            nargs: 1,
            required: false,
            help: 'The folder containing the cartridges, typically /srv/epages/eproot/Cartridges'
        }
    );

    subparser.testcases.addArgument(
        ['-c', '--cartridge'],
        {
            nargs: 1,
            required: false,
            action: 'append',
            help: 'Check if the cartridge has testcases [multi]',
            epilog: 'Example: epnode testcases -c jquery -c presentation'
        }
    );

    // parse arguments
    parseArgs = parser.parseArgs();
    subcommand = parseArgs.subcommand;

    // make it possible to define more than one at a time
    args.profile = parseArgs.profile || undefined;
    args.cartridge = parseArgs.cartridge || undefined;

    // since all arguments are passed as an array, we get'em outa there if we don't need more than one value
    args.storeType = parseArgs.store_type ? parseArgs.store_type[0] : undefined;
    args.version = parseArgs.store_version ? parseArgs.store_version[0] : undefined;
    args.webroot = parseArgs.webroot ? parseArgs.webroot[0] : undefined;
    args.include = parseArgs.include ? parseArgs.include[0] : undefined;
    args.output = parseArgs.output ? parseArgs.output[0] : undefined;
    args.cartridgeFolder = parseArgs.cartridge_folder ? parseArgs.cartridge_folder[0] : undefined;
    args.filter = parseArgs.filter ? parseArgs.filter[0] : undefined;

    async.parallel([
        function (callback) {
            async.series([
                // get environment variables
                function (callback) {
                    perl.get('env', ['webroot', 'appuser', 'webgroup'], function (err, data) {
                        if (!err) {
                            if (data.webroot) {
                                args.webroot = args.webroot || data.webroot;
                                logPath = path.join(data.webroot, '/../Log/');
                                callback();
                            } else {
                                callback("[ERROR] Seems like environment variables are not loaded.");
                            }
                        } else {
                            callback(err.toString());
                        }
                    });
                },

                // use env variables to get uid and gid for unixoids
                function (callback) {
                    if (os.platform() !== 'win32') {
                        async.parallel([
                            function (callback) {
                                perl.get('uid', 'appuser', function (appuser_id) {
                                    uid = appuser_id;
                                    callback();
                                });
                            },

                            function (callback) {
                                perl.get('gid', 'webgroup', function (webgroup_id) {
                                    gid = webgroup_id;
                                    callback();
                                });
                            }
                        ], callback);
                    } else {
                        callback();
                    }
                },

                // packaging
                function (callback) {
                    if (subcommand === 'pack') {
                        if (!(parseArgs.js || parseArgs.css || parseArgs.less)) {
                            console.error("[ERROR] Packaging");
                            callback("Neither JS nor CSS or LESS option provided. Use 'epnode pack --js' | '--css' | '--less");
                            return;
                        }

                        var options = {
                            'storeType': args.storeType,
                            'version': args.version,
                            'webroot': args.webroot,
                            'uid': uid,
                            'gid': gid
                        };

                        async.parallel([
                            function (callback) {
                                if (parseArgs.js) {
                                    async.eachSeries(args.profile, function (profile, callback) {
                                        pack.packJs(profile[0], options, callback);
                                    }, function (error) {
                                        if (error) {
                                            console.error("[ERROR] Packing JS");
                                            buffer = " === Packing JS === \n";
                                            callback(error);
                                        } else {
                                            callback();
                                        }
                                    }, callback);
                                } else {
                                    callback();
                                }
                            },

                            function (callback) {
                                if (parseArgs.css) {
                                    async.eachSeries(args.profile, function (profile, callback) {
                                        pack.packStylesheet(profile[0], options, 'css', callback);
                                    }, function (error) {
                                        if (error) {
                                            console.error("[ERROR] Packing CSS");
                                            buffer = " === Packing CSS === \n";
                                            callback(error);
                                        } else {
                                            callback();
                                        }
                                    }, callback);
                                } else {
                                    callback();
                                }
                            },

                            function (callback) {
                                if (parseArgs.less) {
                                    async.eachSeries(args.profile, function (profile, callback) {
                                        pack.packStylesheet(profile, options, 'less', callback);
                                    }, function (error) {
                                        if (error) {
                                            console.error("[ERROR] Packing LESS");
                                            buffer = " === Packing LESS === \n";
                                            callback(error);
                                        } else {
                                            callback();
                                        }
                                    }, callback);
                                } else {
                                    callback();
                                }
                            }
                        ], callback);
                    } else {
                        callback();
                    }
                },

                // minifying
                function (callback) {
                    if (subcommand === 'minify') {
                        try {
                            var dir = path.join(args.webroot, "StoreTypes", args.version, args.storeType),
                                source = path.join(dir, 'lib'),
                                target = path.join(dir, 'lib-min');

                            minify(
                                source,
                                target,
                                {
                                    "uid": uid,
                                    "gid": gid,
                                    "filter": args.filter
                                },
                                callback
                            );
                        } catch (e) {
                            console.error("[ERROR] Minify");
                            buffer = " === Minifying === \n";
                            callback(e);
                        }
                    } else {
                        callback();
                    }
                }
            ], callback);
        },

        // building documentation
        function (callback) {
            if (subcommand === 'builddoc') {
                args.output = args.output || path.join(process.env.EPAGES_WEBROOT, "Docs", "JavaScript");
                args.cartridgeFolder = args.cartridgeFolder || process.env.EPAGES_CARTRIDGES;

                buildDoc({
                    "input": args.include,
                    "output": args.output,
                    "cartridges": args.cartridgeFolder,
                    "uid": uid,
                    "gid": gid
                }, function (error) {
                    if (error) {
                        console.log("[ERROR] Build documentation");
                    }

                    callback(error);
                });
            } else {
                callback();
            }
        },

        // checking testcases
        function (callback) {
            if (subcommand === 'testcases') {
                check(args.cartridge, function (error) {
                    if (!error) {
                        console.log('Seems like every module is covered by at leat one testcase!');
                        callback();
                    } else {
                        // write buffer for erro if necessary
                        callback(error);

                    }
                });
            } else {
                callback();
            }
        }

    // error handling of all actions
    ], function (error) {
        var stopTime = new Date().getTime(),
            logTime = new Date().toISOString().replace('T', ' ').split('.')[0],
            logMessage,
            errorMessage;

        if (error) {
            //console.log(error.errno, commons.error);
            error.reason = error.errno ? commons.error[error.errno] : undefined;

            // remove the stack trace if extended logging is not activated
            if (!args.extended_log) {
                delete error.stack;
            }

            logMessage = logTime + " " + (buffer || " === General Failure === \n") + JSON.stringify(error, null, 4) + "\n\n";
            /*jslint stupid: true*/
            // To let me write the error log synchronoulsy
            fs.appendFileSync(logPath + "Epnode6.log", logMessage);

            if (error.reason) {
                errorMessage = error.path + ": " + error.reason;
            } else if (error.source) {
                errorMessage = error.source + ": " + error.message + " at line " + error.line + ":" + error.col;
            } else {
                errorMessage = error;
            }

            console.error("[ERROR] " + errorMessage);

            process.exit(1);
        } else {
            console.log("Need " + ((stopTime - startTime) / 1000) + "s");
            process.exit();
        }
    });
}());