/*jslint node:true, nomen:true*/

(function () {
    'use strict';

    var perl = require('./perl'),
        path = require('path'),
        async = require('async'),
        grunt = require('grunt'),
        fs = require('fs'),
        formerWorkingDir = process.cwd();

    module.exports = function (cartridgesToCheck, callback) {

        var missingTestcases = [],
            numberOfModules = 0;

        async.waterfall([
            // prepare cartridges
            function (callback) {
                if (!cartridgesToCheck) {
                    perl.get('cartridges', 'Store', callback);
                } else {
                    callback(null, cartridgesToCheck);
                }

            },

            // do the rest
            function (cartridges, callback) {

                async.each(cartridges, function (cartridge, callback) {

                    // if no cartridge is specified, we let perl get ALL cartridges (format e.g. DE_EPAGES/<CARTRIDGENAME>)
                    var cartridgePath = path.join(process.env.EPAGES_CARTRIDGES, cartridge);

                    async.series([
                        // check if cartridge exists
                        function (callback) {
                            fs.exists(cartridgePath, function (exists) {
                                var err = exists ? null : 'Cannot find cartridge ' + cartridge + '. Format: <NAMESPACE>/<CARTRIDGENAME>';
                                callback(err);
                            });
                        },

                        function (callback) {
                            var jsPath = path.join(cartridgePath, 'Data/Public/lib/');

                            // check if cartridge has a javascript folder (typically ~/Data/Public/lib)
                            fs.exists(jsPath, function (exists) {
                                if (exists) {
                                    // cd jspath
                                    grunt.file.setBase(jsPath);

                                    var modules = grunt.file.expand([
                                        'ep/**/*.js',
                                        cartridge.toLowerCase() + '/**/*.js',
                                        'ckeditor/plugins/ep*/*.js',
                                        '!ep/i18n/*',
                                        '!**/*-test.js',
                                        '!**/*-model.js',
                                        '!**/*-collection.js'
                                    ]);

                                    numberOfModules += modules.length;

                                    async.each(modules, function (module, callback) {
                                        try {
                                            /*jslint stupid:true*/
                                            // seems like the only way to make sure we don't get false positives
                                            // because fs.open() might not able to open files that are actually there
                                            fs.openSync(module.replace(/\.js$/, '-test.js'), 'r');
                                        } catch (e) {
                                            // collect all failing test cases to call'em back at the end, gonna be legen... just wait for it
                                            missingTestcases.push(module);
                                        }

                                        callback();
                                    }, callback);

                                    // cd back
                                    grunt.file.setBase(formerWorkingDir);

                                } else if (cartridges.length > 0 && cartridgesToCheck) {
                                    callback('Cartridge ' + cartridge + ' contains no JavaScript!');
                                } else {
                                    callback();
                                }
                            });
                        }
                    ], callback);
                }, callback);
            }
        ], function (error) {
            var numberOfmissingTestcases = String(missingTestcases.length),
                ratio,
                result = "\n",
                color;

            if (!error) {
                if (numberOfmissingTestcases > 0) {
                    // we have some missing testcases
                    ratio = (numberOfModules - numberOfmissingTestcases) / numberOfModules;

                    if (ratio > 0.75 && ratio !== 1) {
                        color = 'yellow';
                    } else if (ratio > 0.5 && ratio < 0.75) {
                        color = 'magenta';
                    } else {
                        color = 'red';
                    }

                    async.eachSeries(missingTestcases, function (testcase, callback) {
                        // ...still waiting...
                        result += "\t[X] ".red + "Missing test case for " + testcase.cyan + "\n";
                        callback();
                    }, function () {
                        result += "\n==> We have " + numberOfmissingTestcases[color] + " missing test cases | "
                                + numberOfModules + " modules checked | "
                                + String(ratio * 100).substr(0, 5)[color] + " percent coverage.";

                        // ...dary!
                        callback(result);
                    });
                } else {
                    callback();
                }
            } else {
                callback(error);
            }
        });
    };
}());