/*jslint nomen:true,regexp:true,node:true,unparam:true*/
(function () {

    'use strict';

    var commons = require('./commons'),
        uglify = require('uglify-js'),
        grunt = require('grunt'),
        path = require('path'),
        less = require('less'),
        async = require('async'),
        fs = require('fs'),
        PERMISSION = '640', /*r-xr-----*/

        /**
         * Resolve all imports of the file [source] and concatenate the imported files. The imports
         * are resolved recursively, which means that the resulting content must not contain any
         * relative imports.
         *
         * Absolute imports are kept (http://example/style.css and /style.css).
         *
         * @param  {String} source The absolute path to the source file.
         * @return {String}        The content of the source file, containing the resolved imports.
         */
        resolveImports = function (source, base) {
            var content = grunt.file.read(source),
                importRegExp = /@import\s+url\s*\(\s*(["'])?(.*?)\1?\s*\)/g,
                excludeUrlRegExp = /^(data:|[a-z]+:\/\/|\/)/;

            content = content.replace(/(url\s*\(\s*(["']?))(.*?)(\2\s*\))/g, function (fullMatch, beforeUrl, c, url, afterUrl) {
                if (excludeUrlRegExp.test(url)) {
                    // we don't want to modify absolute url, hence we replace the match with itself
                    return fullMatch;
                }

                // we want to make the url, which is currently relative to `path.basename(source)`,
                // relative to the given base directory. to do this, we join the relative path from
                // `base` to `path.dirname(source)` with the url.
                return beforeUrl +
                    commons.fixPath(path.join(path.relative(base, path.dirname(source)), url)) +
                    afterUrl;
            });

            (content.match(importRegExp) || []).forEach(function (i) {
                var importSource = i.replace(importRegExp, '$2');

                if (!excludeUrlRegExp.test(importSource)) {
                    // `importSource` is relative to `base`, because we changed it in the previous
                    // step together with all other urls.
                    importSource = path.join(base, importSource);
                    content = content.replace(i, resolveImports(importSource, base));
                }
            });

            return content;
        };

    // Copy / minify files by type
    module.exports = function (base, source, target, callback) {
        var ext = path.extname(source).slice(1),
            fileName,
            sourceMapFile,
            content,
            parser,
            minified,
            includedFiles;

        if (base.slice(-1) !== "/") {
            base += "/";
        }

        if (ext === 'js') {
            fileName = path.basename(source, "." + ext);

            // contains js and compiled templates
            includedFiles = commons.compileTemplateFiles(base, source, true).includedFiles;

            // Minify Javascript
            minified = uglify.minify(includedFiles, {
                "outSourceMap": fileName + ".js"
            });

            // The minified code needs a link to the source map, which is appended in this step.
            sourceMapFile = path.join(path.dirname(target), fileName + ".js.map");

            async.parallel([
                // We now have got the minfied source as well as the source map. The minified source
                // can be written instantly
                function (callback) {
                    minified.code += "\n//@ sourceMappingURL=" + path.basename(sourceMapFile);
                    grunt.file.write(target, minified.code);

                    fs.chmod(target, PERMISSION, callback);
                },

                // The source map needs some modification. The sources must be relative to the source
                // map.
                function (callback) {
                    minified.map = JSON.parse(minified.map);

                    minified.map.sources.forEach(function (e, i) {
                        var file = path.basename(e),
                            sourceDir = path.dirname(source),
                            targetDir = path.dirname(target);

                        minified.map.sources[i] = commons.fixPath(
                            path.relative(targetDir, path.join(sourceDir, file))
                        );
                    });

                    minified.map = JSON.stringify(minified.map);

                    grunt.file.write(sourceMapFile, minified.map);
                    fs.chmod(sourceMapFile, PERMISSION, callback);
                }
            ], callback);
        } else if (ext === 'css' || ext === 'less') {
            // Minify generated LESS/CSS package files
            // There is no debug level 1 for less/css files, so we're just minifying the actual packages
            if (/^package-[a-z]+/.test(path.basename(source))) {
                content = grunt.file.read(source);

                parser = new (less.Parser)({
                    'paths': [path.dirname(source)],
                    'filename': source
                });

                content = resolveImports(source, base);
                parser.parse(content, function (error, tree) {
                    var basename, dirname;

                    if (error) {
                        throw error;
                    }

                    content = tree.toCSS({
                        compress: true
                    });

                    // Ensure file extension to be .css (since we processed .css AND .less files)
                    basename = path.basename(target).replace(/\.(less|css)$/, '.css');
                    dirname = path.dirname(target);
                    target = path.join(dirname, basename);

                    grunt.file.write(target, content);
                    fs.chmod(target, PERMISSION, callback);
                });
            } else {
                callback();
            }
        } else if (ext === 'html') {
            // HTML to single line without indent
            grunt.file.write(target, grunt.file.read(source).replace(/[\n\r]\s*/g, ''));
            fs.chmod(target, PERMISSION, callback);
        } else {
            // Just copy
            grunt.file.copy(source, target);
            fs.chmod(target, PERMISSION, callback);
        }
    };
}());
