/*jslint indent:4,node:true,nomen:true*/
(function () {

    'use strict';

    var _ = require("underscore"),
        commons = require("./commons"),
        fs     = require("fs.extra"),
        path   = require("path"),
        perl   = require("./perl"),
        async  = require("async"),
        PERMISSION = '640',
        getCartridgesFolder,
        getCartridgeFolder,
        getProfiles,
        concatenateProfiles,
        createJsPackageFile,
        createJsPackageInfoFile,
        deleteFileIfExists,
        constructStylePackagePath;

    getCartridgesFolder = function (callback) {
        // Get the cartridge folder from the perl script
        perl.get("env", ["cartridges"], function (error, data) {
            callback(error, data.cartridges);
        });
    };

    /**
     * Returns the containing folder for the given cartridge. Example:
     * "/srv/epages/eproot/Cartridges/DE_EPAGES/JQuery" for the cartridge "DE_EPAGES/JQuery".
     *
     * @param {string}   name     The name of the cartridge, for example "DE_EPAGES/JQuery".
     * @param {string}   basePath The base path which contains the cartridges. In case this
     *                            parameter is omitted, the base path is fetched from the perl
     *                            script.
     * @param {function} callback The function to call when the result is available.
     */
    getCartridgeFolder = function (name, basePath, callback) {
        // If the `callback` is not defined and the `basePath` is a function, the user may want to
        // use the `basePath` as `callback` instead.
        if (!callback && _.isFunction(basePath)) {
            callback = basePath;
            basePath = undefined;
        }

        // The `cfCallback` is called as soon as the `basePath` is available. This function joins
        // the `basePath` and the name of the cartridge and invokes the `callback`.
        var cfCallback = function (error, data) {
            callback(error, path.join(data, name));
        };

        if (!basePath) {
            // The `basePath` is undefined, hence we must invoke the perl script. The `cfCallback`
            // is invoked with the `basePath` as soon as the result is available.
            getCartridgesFolder(cfCallback);
        } else {
            // The user gave us a `basePath`, which we can use instantly.
            cfCallback(null, basePath);
        }
    };

    /**
     * Returns an array of profile arrays for the given cartridges and of the given types. A
     * cartridge may be something like "DE_EPAGES/JQuery". The type may have one of the following
     * values: "bo", "sf", "mobilesf", "editor". The (optional) {@code basePath} points to the
     * folder containing the cartrides. For the cartridge "DE_EPAGES/JQuery", the {@code basePath}
     * must contain the folder "DE_EPAGES".
     *
     * @param {string}   type       The type of the included profiles ("bo", "sf", "mobilesf" or
     *                              "editor").
     * @param {string}   basePath   The base path which contains the cartridges. In case this
     *                              parameter is omitted, the base path is fetched from the perl
     *                              script.
     * @param {[string]} cartridges An array of files to include.
     */
    getProfiles = function (type, cartridges, basePath, callback) {
        // If the `callback` is not defined and the `basePath` is a function, the user may want to
        // use the `basePath` as `callback` instead.
        if (!callback && _.isFunction(basePath)) {
            callback = basePath;
            basePath = undefined;
        }

        // We create an array of functions, where each function loads a *script-package-<profile>.json* for a
        // cartridge. This array is used for an `async.parallel` call in the next step.
        var pendingWaterfalls = [];
        cartridges.forEach(function (cartridge) {
            pendingWaterfalls.push(function (callback) {
                async.waterfall([
                    function (callback) {
                        // We determine where the *script-package-<profile>.json* is located for the current
                        // cartridge.
                        getCartridgeFolder(cartridge, basePath, function (error, folder) {
                            callback(
                                error,
                                path.join(folder, "Data", "Public", "lib", cartridge.toLowerCase(), "script-package-" + type + ".json")
                            );
                        });
                    },

                    function (profileJs, callback) {
                        // And check whether the *script-package-<profile>.json* exists.
                        fs.exists(profileJs, function (exists) {
                            callback(undefined, exists ? profileJs : undefined);
                        });
                    },

                    function (profileJs, callback) {
                        if (profileJs) {
                            // Then we load the JSON array from the *script-package-<profile>.json* if it exists
                            fs.readFile(profileJs, function (error, data) {
                                var currProfile = JSON.parse(data);

                                currProfile.forEach(function (v, idx) {
                                    currProfile[idx] = {
                                        "name": v.replace(/\./g, "/"),
                                        "metaparser": true
                                    };
                                });

                                callback(error, currProfile);
                            });
                        } else {
                            // Or return an undefined value, if it doesn't exist.
                            callback(undefined, undefined);
                        }
                    }
                ], callback);
            });
        });

        // Now we execute the array of functions from the previous step. This should generate an
        // array, where each element either contains the content of a *script-package-<profile>.json* of a cartridge,
        // or `undefined`, if the correpsonding cartridge has no *script-package-<profile>.json*.
        async.parallel(pendingWaterfalls, function (error, result) {
            // Because some elements of the `result` array may be undefined, we clean those emtpy
            // spaces.
            var cleanedProfiles = [];
            result.forEach(function (profile) {
                if (profile) {
                    cleanedProfiles.push(profile);
                }
            });
            // `cleanedProfiles` is an array of all existing *script-package-<profile>.json* files. The order of this
            // array matches the order of the cartridges: If *CartridgeA* is before *CartridgeB*,
            // then the *script-package-<profile>.json* of *CartridgeA* will be before the one of *CartridgeB*.
            callback(error, cleanedProfiles);
        });
    };

    /**
     * Concatenates the elements of the given array of profiles and removes duplicates.
     *
     * @param  {[[{}]]} profiles An array of profiles.
     * @return {[{}]}            The concatenated array.
     */
    concatenateProfiles = function (profiles) {
        // We concatenate the profiles...
        var concatenatedProfiles = [];
        profiles.forEach(function (profile) {
            concatenatedProfiles = concatenatedProfiles.concat(profile);
        });

        // ...and remove duplicates. The strategy is to keep the first occurence and drop all
        // following occurences. Two entries `a` and `b` are considered as being the same if
        // `a.name == b.name`
        concatenatedProfiles = _.uniq(concatenatedProfiles, function (o) {
            return o.name;
        });

        return concatenatedProfiles;
    };

    deleteFileIfExists = function (file, callback) {
        async.waterfall([
            function (callback) {
                // We check if the file exists...
                fs.exists(file, function (exists) {
                    callback(null, exists);
                });
            },

            function (exists, callback) {
                if (exists) {
                    // ...and remove it if it exists
                    fs.unlink(file, function (error) {
                        callback(error);
                    });
                } else {
                    // ...or do nothing if it doesn't exist
                    callback(null);
                }
            }
        ], function (error) {
            callback(error);
        });
    };

    /**
     * Creates a package.js from the individual JavaScript files. The files to include are listed
     * in the array {@code concatenatedProfiles}.
     *
     * @param  {[{}]}     concatenatedProfiles A list of all files to include in the package.
     * @param  {function} callback             A callback function which is called if the package.js
     *                                         is available.
     */
    createJsPackageFile = function (concatenatedProfiles, profile, targetPath, callback) {
        var target = path.join(targetPath, "package-" + profile + ".js");
        concatenatedProfiles.unshift({
            name: "package-" + profile + ".info",
            metaparser: false
        });

        async.series([
            // remove existing *package-??.js*
            function (callback) {
                deleteFileIfExists(target, callback);
            },

            function (callback) {
                var pendingWaterfalls = [];
                concatenatedProfiles.forEach(function (file) {
                    var source = path.join(targetPath, file.name + ".js");

                    // read content and append to *package-??.js*
                    pendingWaterfalls.push(function (callback) {
                        async.waterfall([
                            function (callback) {
                                var jsWithTemplates = commons.compileTemplateFiles(targetPath, source, false).concatenatedContent;
                                callback(null, jsWithTemplates);
                            },

                            function (fileContent, callback) {
                                fs.appendFile(target, fileContent, callback);
                            }
                        ], callback);
                    });
                });

                async.series(pendingWaterfalls, callback);
            }
        ], callback);
    };

    createJsPackageInfoFile = function (includedFiles, profile, targetPath, callback) {
        var target = path.join(targetPath, "package-" + profile + ".info.js");

        async.waterfall([
            function (callback) {
                async.map(includedFiles, function (e, callback) {
                    callback(null, e.metaparser ? e.name : null);
                }, callback);
            },

            function (includedFiles, callback) {
                async.reduce(includedFiles, [], function (memo, item, callback) {
                    if (item) {
                        memo.push(item);
                    }

                    callback(null, memo);
                }, callback);
            },

            function (includedFiles, callback) {
                fs.writeFile(target, 'var require=function(require){var metaparseConfig=require||(require={});metaparseConfig=require.config||(require.config={});metaparseConfig=metaparseConfig["jquery/metaparse"]||(metaparseConfig["jquery/metaparse"]={});metaparseConfig.autoparse={selector:metaparseConfig.autoparse?".ep-js":"#PREVENT-METAPARSE",require:epkStack=' + JSON.stringify(includedFiles) + '};return require}(this.require);', callback);
            }
        ], callback);
    };

    /**
     * Creates a package of JS files. Which files are included in this package can be controlled
     * by the {@code config} parameter.
     *
     *     {@code config.storeType}  is the store type the packaged CSS file is for.
     *     {@code config.profile}    is the profile (bo, sf, mobilesf or editor) the packaged CSS
     *                               file is for.
     *     {@code config.webroot}    is the webroot, which containts the store types.
     *     {@code config.version}    is the store version for which the packaged file is created.
     *     {@code config.uid}        is the id of the user which will own the created files. This
     *                               option is optional. The uid of the created files will only be
     *                               changed if the gid is given.
     *     {@code config.gid}        is the id of hte group to which the created files will belong.
     *                               This option is optional. The gid of the created files will only
     *                               be changed if the uid is given.
     *
     * @param  {Object}   config   The configuration. See the documentation comment for further
     *                             details.
     * @param  {Function} callback The callback function which is called as soon as the package file
     *                             is available. It can take on argument, which is the occured
     *                             error.
     */
    module.exports.packJs = function (profile, config, callback) {
        async.waterfall([
            function (callback) {
                if (!config.targetPath) {
                    config.targetPath = path.join(config.webroot, "StoreTypes", config.version, config.storeType, "lib");
                }

                callback(null, config.targetPath);
            },

            function (targetPath, callback) {
                if (!config.sourcePath) {
                    // If `config.sourcePath` is not set, we take it from the perl script.
                    getCartridgesFolder(function (error, sourcePath) {
                        callback(error, sourcePath, targetPath);
                    });
                } else {
                    // If `config.sourcePath` is set, we use it.
                    callback(null, config.sourcePath, targetPath);
                }
            },

            function (sourcePath, targetPath, callback) {
                perl.get("cartridges", config.storeType, function (error, cartridges) {
                    if (cartridges) {
                        callback(error, sourcePath, targetPath, cartridges);
                    } else {
                        callback("No cartridges found for the store type " + config.storeType);
                    }
                });
            },

            function (sourcePath, targetPath, cartridges, callback) {
                getProfiles(profile, cartridges, sourcePath, function (error, profiles) {
                    callback(error, profiles, targetPath);
                });
            },

            function (profiles, targetPath, callback) {
                var includedFiles = concatenateProfiles(profiles);

                async.series([
                    function (callback) {
                        createJsPackageInfoFile(includedFiles, profile, targetPath, callback);
                    },

                    function (callback) {
                        createJsPackageFile(includedFiles, profile, targetPath, callback);
                    }
                ], function (error) {
                    callback(error, targetPath);
                });
            },

            function (targetPath, callback) {
                var targetInfoFile = path.join(targetPath, "package-" + profile + ".info.js"),
                    targetJsFile = path.join(targetPath, "package-" + profile + ".js");

                if (_.isNumber(config.uid) && _.isNumber(config.gid)) {
                    async.parallel([
                        function (callback) {
                            fs.chown(targetInfoFile, config.uid, config.gid, callback);
                        },

                        function (callback) {
                            fs.chmod(targetInfoFile, PERMISSION, callback);
                        },

                        function (callback) {
                            fs.chown(targetJsFile, config.uid, config.gid, callback);
                        },

                        function (callback) {
                            fs.chmod(targetJsFile, PERMISSION, callback);
                        }
                    ], callback);
                } else {
                    callback();
                }
            }
        ], function (error) {
            callback(error);
        });
    };

    /**
     * This function constructs the (absolute) path to a style package. The path is constructed as
     * follows:
     *
     * stylePackagePath = [targetPath]/[middle]/style-package-[profile][extensionPrefix].css
     *
     *  @param  {string} targetPath         The first part, an absolute path.
     * @param  {string} middle             The parent directory of the style package, relative to
     *                                     the [targetPath].
     * @param  {string} profile            The profile, like "bo" or "sf".
     * @param  {string} [extensionPrefix=] A string which prepends the extension. Typically
     *                                     "-variables".
     * @param  {string} [extension=css]    The file extension.
     * @return {string}                    The absolute path to the style package.
     */
    constructStylePackagePath = function (targetPath, middle, profile, extensionPrefix, extension) {
        if (!extensionPrefix) {
            extensionPrefix = "";
        }

        if (!extension) {
            extension = "css";
        }

        return path.join(
            targetPath,
            middle,
            "style-package-" + profile + extensionPrefix + "." + extension
        );
    };

    /**
     * Creates a package of CSS files. Which files are included in this package can be controlled
     * by the {@code config} parameter.
     *
     *     {@code config.storeType}  is the store type the packaged CSS file is for.
     *     {@code config.profile}    is the profile (bo, sf, mobilesf or editor) the packaged CSS
     *                               file is for.
     *     {@code config.targetPath} is the target path for the packaged CSS file and the source
     *                               path for the individual parts of the package. This option is
     *                               optional. If it is not set, the correct target path is created
     *                               automatically.
     *     {@code config.uid}        is the id of the user which will own the created files. This
     *                               option is optional. The uid of the created files will only be
     *                               changed if the gid is given.
     *     {@code config.gid}        is the id of hte group to which the created files will belong.
     *                               This option is optional. The gid of the created files will only
     *                               be changed if the uid is given.
     *
     * @param  {{}}       config   The configuration. See the documentation comment for further
     *                             details.
     * @param  {function} callback The callback function which is called as soon as the package file
     *                             is available. It can take on argument, which is the occured
     *                             error.
     */
    module.exports.packStylesheet = function (profile, config, fileExtension, callback) {
        async.waterfall([
            function (callback) {
                // We get the cartridges to include for the requested store type. This gives us the
                // order of the cartridges at the same time.
                perl.get("cartridges", config.storeType, function (error, cartridges) {
                    if (cartridges) {
                        callback(error, cartridges);
                    } else {
                        callback("No cartridges found for the store type " + config.storeType);
                    }
                });
            },

            function (cartridges, callback) {
                if (!config.targetPath) {
                    config.targetPath = path.join(config.webroot, "StoreTypes", config.version, config.storeType, "lib");

                }

                callback(null, config.targetPath, cartridges);
            },

            function (targetPath, cartridges, callback) {
                var paths = [];

                // We add the paths for *~/lib/[cartridge]* (less)
                cartridges.forEach(function (cartridge) {
                    paths.push(constructStylePackagePath(targetPath, cartridge.toLowerCase(), profile, "-variables", "less"));
                });

                // We add the paths for *~/lib/[cartridge]* (css and less)
                cartridges.forEach(function (cartridge) {
                    paths.push(constructStylePackagePath(targetPath, cartridge.toLowerCase(), profile));
                    paths.push(constructStylePackagePath(targetPath, cartridge.toLowerCase(), profile, "", "less"));
                });

                callback(null, targetPath, paths);
            },

            function (targetPath, paths, callback) {
                // Some of the paths created in the previous step don't exist. We don't need them
                // anymore, hence we drop them.
                async.filter(paths, fs.exists, function (paths) {
                    callback(null, targetPath, paths);
                });
            },

            function (targetPath, paths, callback) {
                var content = "",
                    targetFile;

                // We build the targetFile, which is located under targetPath (*~/lib*) and named
                // after the schema *package-[profile].css*.
                targetFile = path.join(targetPath, "package-" + profile + "." + fileExtension);

                // concatenate the import statements.
                paths.forEach(function (p) {
                    p = commons.fixPath(path.relative(targetPath, p));
                    content += "@import url(\"" + p + "\");\n";
                });

                callback(null, targetFile, content);
            },

            function (targetFile, content, callback) {
                // We now write the *package-[profile].css* file into [targetFile].
                fs.writeFile(targetFile, content, function (error) {
                    callback(error, targetFile);
                });
            },

            function (targetFile, callback) {
                // if on unix, set file ownership and permisions
                if (_.isNumber(config.uid) && _.isNumber(config.gid)) {
                    async.parallel([
                        function (callback) {
                            fs.chown(targetFile, config.uid, config.gid, callback);
                        },

                        function (callback) {
                            fs.chmod(targetFile, PERMISSION, callback);
                        }
                    ], callback);
                } else {
                    callback();
                }
            }
        ], function (error) {
            callback(error);
        });
    };

}());