/*jslint node:true*/
/*global describe,before,it,after*/
(function () {

    'use strict';

    var async = require("async"),
        commonsTesting = require('./commons-testing'),
        expect = require('expect.js'),
        fs = require("fs"),
        grunt = require("grunt"),
        sinon = require("sinon"),
        minify = require("../minify-batch"),
        pack = require("../pack"),
        path = require("path"),
        os = require('os'),
        perl = require('../perl'),

        is_unixWithRoot = (os.platform() !== 'win32' && process.getuid() === 0) ? 'Running on Unix with root, performing checks for uid and gid' : undefined;


    describe("Running pack.js and minify.js", function () {

        if (is_unixWithRoot) {
            console.log(is_unixWithRoot);
        }

        before(function (callback) {
            // mock the perl module
            sinon.stub(perl, 'get').withArgs('cartridges', 'Site').yields(null, [
                "PARENT/JQuery",
                "PARENT/CartridgeD",
                "PARENT/CartridgeE",
                "PARENT/CartridgeB",
                "Another_Parent/CartridgeC",
                "PARENT/CartridgeA"
            ]);

            callback();
        });

        describe("pack", function () {
            describe("--less and --css", function () {
                var actualPath   = commonsTesting.getPath("webroot-lib"),
                    expectedPath = commonsTesting.getPath("expected"),
                    options = {
                        "targetPath": actualPath,
                        "storeType": "Site"
                    };

                if (is_unixWithRoot) {
                    options.uid = 1337;
                    options.gid = 42;
                }

                it("should create and write the correct content into package-bo.less", function (callback) {
                    pack.packStylesheet('bo', options, 'less', function (error) {
                        // no error should be reported
                        expect(error).to.not.be.ok();

                        async.parallel([
                            function (callback) {
                                // package-bo.less should exist
                                fs.exists(path.join(actualPath, "package-bo.less"), function (exists) {
                                    expect(exists).to.be.ok();
                                    callback();
                                });
                            },

                            function (callback) {
                                // contents of actual package-bo.less should match content of expected package-bo.less
                                async.parallel({
                                    expected: function (callback) {
                                        fs.readFile(path.join(expectedPath, "package-bo.less"), callback);
                                    },

                                    actual: function (callback) {
                                        fs.readFile(path.join(actualPath, "package-bo.less"), callback);
                                    }
                                }, function (error, result) {
                                    expect(error).to.not.be.ok();
                                    expect(result.expected.toString()).to.eql(result.actual.toString());

                                    // if we're root, we check if file ownership is set correctly
                                    if (is_unixWithRoot) {
                                        // check uid, gid and permissions
                                        fs.stat(path.join(actualPath, "package-bo.less"), function (err, stats) {
                                            expect(err).to.not.be.ok();

                                            expect(stats.uid).to.be.a('number');
                                            expect(stats.uid).to.be(1337);
                                            expect(stats.gid).to.be.a('number');
                                            expect(stats.gid).to.be(42);

                                            expect(parseInt(stats.mode.toString(8), 10)).to.be(100640);

                                            callback();
                                        });
                                    } else {
                                        callback();
                                    }
                                });
                            }
                        ], callback);
                    });
                });
            });

            describe("--js", function () {
                var cartridgePath = commonsTesting.getPath("cartridges"),
                    actualPath    = commonsTesting.getPath("webroot-lib"),
                    expectedPath  = commonsTesting.getPath("expected"),
                    options = {
                        "sourcePath": cartridgePath,
                        "targetPath": actualPath,
                        "storeType": "Site"
                    };

                if (is_unixWithRoot) {
                    options.uid = 1337;
                    options.gid = 42;
                }

                it("should create and write the correct content into package-bo.js and package-bo.info.js", function (callback) {
                    pack.packJs('bo', options, function (error) {
                        // no error should be reported
                        expect(error).to.not.be.ok();

                        async.parallel([
                            function (callback) {
                                // package-bo.js should exist
                                fs.exists(path.join(actualPath, "package-bo.js"), function (exists) {
                                    expect(exists).to.be.ok();
                                    callback();
                                });
                            },

                            function (callback) {
                                // contents of actual package-bo.js should match content of expected package-bo.js
                                async.parallel({
                                    expected: function (callback) {
                                        fs.readFile(path.join(expectedPath, "package-bo.js"), callback);
                                    },

                                    actual: function (callback) {
                                        fs.readFile(path.join(actualPath, "package-bo.js"), callback);
                                    }
                                }, function (error, result) {
                                    expect(error).to.not.be.ok();
                                    expect(result.expected.toString()).to.eql(result.actual.toString());

                                    // if we're root, we check if file ownership is set correctly
                                    if (is_unixWithRoot) {
                                        // check uid, gid and permissions
                                        fs.stat(path.join(actualPath, "package-bo.js"), function (err, stats) {
                                            expect(err).to.not.be.ok();

                                            expect(stats.uid).to.be.a('number');
                                            expect(stats.uid).to.be(1337);
                                            expect(stats.gid).to.be.a('number');
                                            expect(stats.gid).to.be(42);

                                            expect(parseInt(stats.mode.toString(8), 10)).to.be(100640);

                                            callback();
                                        });
                                    } else {
                                        callback();
                                    }
                                });
                            },

                            function (callback) {
                                // contents of actual package-bo.info.js should match content of expected
                                // package-bo.info.js
                                async.parallel({
                                    expected: function (callback) {
                                        fs.readFile(path.join(expectedPath, "package-bo.info.js"), callback);
                                    },

                                    actual: function (callback) {
                                        fs.readFile(path.join(actualPath, "package-bo.info.js"), callback);
                                    }
                                }, function (error, result) {
                                    expect(error).to.not.be.ok();
                                    expect(result.expected.toString()).to.eql(result.actual.toString());

                                    // if we're root, we check if file ownership is set correctly
                                    if (is_unixWithRoot) {
                                        // check uid, gid and permissions
                                        fs.stat(path.join(actualPath, "package-bo.info.js"), function (err, stats) {
                                            expect(err).to.not.be.ok();

                                            expect(stats.uid).to.be.a('number');
                                            expect(stats.uid).to.be(1337);
                                            expect(stats.gid).to.be.a('number');
                                            expect(stats.gid).to.be(42);

                                            expect(parseInt(stats.mode.toString(8), 10)).to.be(100640);

                                            callback();
                                        });
                                    } else {
                                        callback();
                                    }
                                });
                            }
                        ], callback);
                    });
                });
            });
        });

        describe("minify", function () {
            it("should minify the files and resolve @imports only in the ^package-* files", function (callback) {
                var webrootLib    = commonsTesting.getPath("webroot-lib"),
                    webrootLibMin = commonsTesting.getPath("webroot-lib-min"),
                    expectedPath  = commonsTesting.getPath("expected"),
                    options = {};

                if (is_unixWithRoot) {
                    options.uid = 1337;
                    options.gid = 42;
                }

                async.series([
                    function (callback) {
                        minify(webrootLib, webrootLibMin, options, callback);
                    },

                    function (callback) {
                        async.parallel([
                            function (callback) {
                                // contents of actual a2.js.map should match content of expected a2.js.map
                                async.parallel({
                                    expected: function (callback) {
                                        fs.readFile(path.join(expectedPath, "parent-a2.min.js.map"), callback);
                                    },

                                    actual: function (callback) {
                                        fs.readFile(path.join(webrootLibMin, "parent", "a2.js.map"), callback);
                                    }
                                }, function (error, result) {
                                    expect(error).to.not.be.ok();
                                    expect(result.expected.toString()).to.be(result.actual.toString());

                                                                        // if we're root, we check if file ownership is set correctly
                                    if (is_unixWithRoot) {
                                        // check uid, gid and permissions
                                        fs.stat(path.join(webrootLibMin, "parent", "a2.js.map"), function (err, stats) {
                                            expect(err).to.not.be.ok();

                                            expect(stats.uid).to.be.a('number');
                                            expect(stats.uid).to.be(1337);
                                            expect(stats.gid).to.be.a('number');
                                            expect(stats.gid).to.be(42);

                                            expect(parseInt(stats.mode.toString(8), 10)).to.be(100640);

                                            callback();
                                        });
                                    } else {
                                        callback();
                                    }
                                });
                            },

                            function (callback) {
                                // contents of actual minifed package-bo.css should match content of
                                // expected package-bo.css
                                async.parallel({
                                    expected: function (callback) {
                                        fs.readFile(path.join(expectedPath, "package-bo.min.css"), callback);
                                    },

                                    actual: function (callback) {
                                        fs.readFile(path.join(webrootLibMin, "package-bo.css"), callback);
                                    }
                                }, function (error, result) {
                                    expect(error).to.not.be.ok();
                                    expect(result.expected.toString()).to.be(result.actual.toString());

                                    // if we're root, we check if file ownership is set correctly
                                    if (is_unixWithRoot) {
                                        // check uid, gid and permissions
                                        fs.stat(path.join(webrootLibMin, "package-bo.css"), function (err, stats) {
                                            expect(err).to.not.be.ok();

                                            expect(stats.uid).to.be.a('number');
                                            expect(stats.uid).to.be(1337);
                                            expect(stats.gid).to.be.a('number');
                                            expect(stats.gid).to.be(42);

                                            expect(parseInt(stats.mode.toString(8), 10)).to.be(100640);

                                            callback();
                                        });
                                    } else {
                                        callback();
                                    }
                                });
                            },

                            function (callback) {
                                // Expect template files not to be in the minfied directory
                                fs.exists(path.join(webrootLibMin, "backbone-someview.tmpl.html"), function (exists) {
                                    expect(exists).to.not.be.ok();
                                    callback();
                                });
                            },

                            function (callback) {
                                // Expect compiled templates not to be in the source directory
                                fs.exists(path.join(webrootLib, "backbone-someview.tmpl.js"), function (exists) {
                                    expect(exists).to.not.be.ok();
                                    callback();
                                });
                            }
                        ], function () {
                            callback();
                        });
                    }
                ], callback);
            });
        });

        after(function (callback) {
            // clean produced files
            async.parallel([
                function (callback) {
                    var pending = [];
                    grunt.file.expand(path.join(commonsTesting.getPath("webroot-lib"), "package-*")).forEach(function (p) {
                        pending.push(function (callback) {
                            fs.unlink(p, callback);
                        });
                    });

                    async.parallel(pending, callback);
                },

                function (callback) {
                    fs.exists(commonsTesting.getPath("webroot-lib-min"), function (e) {
                        if (e) {
                            grunt.file["delete"](commonsTesting.getPath("webroot-lib-min"));
                        }

                        callback();
                    });
                },

                function (callback) {
                    // restore perl module
                    perl.get.restore();

                    callback();
                }
            ], callback);
        });
    });

}());
